//---------------------------------------------------------------------------
// Copyright 2003 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/corr/main.c,v $
// $Id: main.c,v 1.5 2004/04/21 21:44:11 ah Exp $ 
//
// Description: correlation of files
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: main.c,v 1.5 2004/04/21 21:44:11 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <stdlib.h>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Config/Types.h>
#include <AH/Etech/Types.h>
#include <AH/Etech/Riaa.h>
#include <AH/Etech/Algo.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

#include "corr.h"

// --- locaL definitions ---

// --- local constants ---

static const char* cmdLineOptions = "Cc:dhnr:v";
static const char* version = "0.7";

// --- local data ---

static unsigned int debon = 0;
static bool normalize = false;
static bool cyclic = false;
static unsigned int channel = 1;

//static char mode = '?';

// --- local classes ---

// --- local functions ---


static void usage()
{
    cerr << "Usage: corr [-" << cmdLineOptions << 
	"] inputfile [inputfile] outputfile\n"
	 << "       make an auto or cross correlation.\n"
	 << "  options:\n"
	 << "    h         - print this help message\n"
	 << "    C         - cyclic correlation\n"
	 << "    c <num>   - channel number (for multi channel WAV files)\n"
	 << "                stereo: 1=left 2=right (default=1)\n"
	 << "    d         - increase debug output (may be given more than once)\n"
	 << "    n         - normalize to highest point = 1 (not for WAV output)\n"
	 << "    v         - print version\n"
	 << "  input file  - WAV file or 16 bit mono PCM\n"
	 << "  output file - text file for gnuplot or mono WAV file" << endl;
} // end of usage()



static void print_version()
{
    cerr << "corr Vs " << version << " " 
	 << "(c) 2003 Andreas Hnnebeck\n"
	 << "Credits to Patrick Piecha and Thomas Ludwig.\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()



static unsigned int read_pcm_file(int32* data, unsigned int num, FILE* fpi)
{
    if (debon > 1) cerr << "read_file: read " << num << " points" << endl;
    unsigned int i = 0;

    int ret = 0;
    do
    {
	const unsigned int snum = 65536;
	int16 sbuf[snum];

	// set size of data to read on this turn:
	unsigned int rnum = snum;
	if (rnum > num) rnum = num;

	// read input buffer:
	if (debon > 2) cerr << "read_file: read " << rnum << " points" << endl;
	ret = fread(sbuf, sizeof(int16), rnum, fpi);
	if (debon > 2) cerr << "read_file: read returns " << ret << endl;
	if (-1 == ret) throw ErrorSys();
	for (int k = 0; k < ret; ++k)
	    data[i++] = sbuf[k];
	num -= ret;
	if ((unsigned int)ret < rnum) break;
    } while (num > 0);

    if (debon > 1) cerr << "read_file: returns " << i << endl;
    return i; // number of read entries
}



static uint32 read_wav_file(int32* data, uint32 num, IORiffWaveFile& wave)
{
    if (debon > 1) cerr << "read_file: read " << num << " points" << endl;
    uint32 i = 0;


    // get size of data for desired samples
    const uint32 snum = 16384; // number of samples to read
    const uint32 bufsize = wave.getDataBufSize(snum);

    int ret = 0;
    do
    {
	unsigned char sbuf[bufsize];
	memset(sbuf, 0, bufsize);

	// set size of data to read on this turn:
	uint32 rnum = snum;
	uint32 rbufsize = bufsize;
	if (rnum > num)
	{
	    rnum = num;
	    rbufsize = wave.getDataBufSize(num);
	}

	// read input buffer:
	if (debon > 2) cerr << "read_file: read " << rnum << " points" << endl;
	ret = wave.readRaw(sbuf, rbufsize);
	if (ret <= 0) break;
	if ((uint32)ret < rbufsize)
	    cerr << "Warning: could not read " << (rbufsize - ret) 
		 << " bytes from WAV file" << endl;  
	for (uint32 k = 0; k < rnum; ++k)
	{
	    data[i++] = wave.getBufSample(sbuf, k, channel);
	}
	num -= rnum;
    } while (num > 0);

    if (debon > 1) cerr << "read_file: returns " << i << endl;
    return i; // number of read entries
}



bool isWaveFile(const std::string& filename)
{
    const char* fn = filename.c_str();
    int len = strlen(fn);
    if (0 == strcasecmp(".wav", fn + len - 4))
	return true;
    return false;
}



// get normalization factor 
static double getNormFact(uint32 num, const corr_out_t* data)
{
    if (!normalize) return 1.0;

    // normalize at highest point
    if (debon) cerr << "normalize data at highest point" << endl;
    corr_out_t max = 0;
    corr_out_t min = 0;
    for (unsigned int i = 0; i < num; ++i)
    {
	corr_out_t val = data[i];
	if (val > max) max = val;
	if (val < min) min = val;
    }
    min = -min;
    if (min > max) max = min;
    if (max == 0) return 1.0;

    double norm_fact = 1.0 / max;
    if (debon > 1) 
	cerr << "normalize data : peak = " << max << " norm_fact = "
	     << norm_fact << endl;
    return norm_fact;
}



template <class corr_t> static void calc_stat(
    const std::string&  head, corr_t* data, unsigned int num)
{
    // determine min, mean and max value:
    corr_t mean = 0;
    corr_t max = 0;
    corr_t min = 0;
    for (unsigned int i = 0; i < num; ++i)
    {
	mean += data[i];
	if (data[i] > max) max = data[i];
	if (data[i] < min) min = data[i];
    }
    mean /= num;
    cerr << head << ": min = " << min << ", mean = " << mean 
	 << ", max = " << max << "; num=" << num << endl;
}



static int32* read_file
(
    const string& in,
    uint32& num,      // number of samples
    uint32& sr,       // sample rate in Hz
    uint32& sb        // number of sample bits
)
{
    //------------------------------------
    // read the data from the input file:
    //------------------------------------

    int32* din = 0;      // input samples (int32)
    if (isWaveFile(in))
    {
	IORiffWaveFile wave(in.c_str(), false, debon);
	if (debon > 1) wave.printInfo(cerr);
	uint32 chnum = wave.getChannelNum();
	if (channel > chnum) 
	    throw ErrorBasic("illegal argument following option -c");

	num = wave.getSampleNum();
	sb = wave.getSampleSize();
	uint32 sr_in = wave.getSampleRate();

	if (0 == sr) sr = sr_in;
	else if (sr != sr_in) throw ErrorBasic("non matching samplerate");

	// create input buffer
	din = new int32[num];
	if (!din) throw ErrorSysOutOfMem();

	// read input data into input buffer:
	if (debon) cerr << "Input: reading wav file " << in << endl;
	uint32 gotten = read_wav_file(din, num, wave);
	if (((int)num - (int)gotten) > 0)
	{
	    cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << endl;
	    num = gotten;
	}
    }
    else // binary PCM file
    {
	// get size of file:
	struct stat st;
	int ret = stat(in.c_str(), &st);
	if (0 != ret) throw ErrorSys("Cannot stat " + in);

	// determine number of data points:
	num = st.st_size / sizeof(short);
	if (debon > 2) cerr << in << ": size=" << st.st_size << " = " << num
			    << " points" << endl;
	if (2 > num) throw ErrorBasic("Size of " + in + " is too small");

	// open data file for read:
	FILE* fpi = fopen(in.c_str(), "rb");
	if (!fpi) throw ErrorSys("Cannot open " + in);

	// create input buffer
	din = new int32[num];
	if (!din)
	{
	    fclose(fpi);
	    throw ErrorSysOutOfMem();
	}

	// read input data into input buffer:
	if (debon) cerr << "Input: reading PCM file " << in << endl;
	uint32 gotten = read_pcm_file(din, num, fpi);

	// close file, done:
	fclose(fpi);
	if (((int)num - (int)gotten) > 0)
	{
	    cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << endl;
	    num = gotten;
	}

    }
    if (debon) calc_stat (in, din, num);
    return din;
}


static void write_file(const string& out, uint32 sr, uint32 sb, 
		       corr_out_t* data, uint32 num)
{
    if (isWaveFile(out))
    {
	// create WAV file:
	int channels = 1;
	IORiffWaveFile wave(out.c_str(), channels, sb, sr, num);
    
	// get size of data for desired samples
	const uint32 snum = 65536; // number of samples to write
	const uint32 bufsize = wave.getDataBufSize(snum);

	// we MUST normalize the data to fit into 'sb' bits:
	corr_out_t min = 0;
	corr_out_t max = 0;
	for (unsigned int k = 0; k < num; ++k)
	{
	    corr_out_t cur = data[k];
	    if (cur > max) max = cur;
	    if (cur < min) min = cur;
	}
	min = -min;
	if (max < min) max = min;
	double norm = (double)wave.getMaxAmplitude()/(double)max;

	uint32 osize = num;
	uint32 i = 0;
	int ret = 0;
	do
	{
	    unsigned char sbuf[bufsize];
	    memset(sbuf, 0, bufsize);

	    // set size of data to read on this turn:
	    uint32 rnum = snum;
	    uint32 rbufsize = bufsize;
	    if (rnum > osize)
	    {
		rnum = osize;
		rbufsize = wave.getDataBufSize(osize);
	    }

	    // write corr_out_t data (rounded) into output buffer
	    for (uint32 k = 0; k < rnum; ++k)
	    {
		double db_mean = norm * data[i];
		int32 val = (int32)(db_mean + 0.5);
		if (debon > 3)
		    cerr << "write_file: val = " << val << endl;
		for (int chan = 0; chan < channels; ++chan)
		    wave.setBufSample(sbuf, k, chan+1, val);
		++i;
	    }

	    // write output buffer:
	    if (debon > 2) 
		cerr << "write_file: write " << rnum << " points" << endl;
	    ret = wave.writeRaw(sbuf, rbufsize);
	    if (ret <= 0) break;
	    if ((uint32)ret < rbufsize)
		cerr << "Warning: could not write " << (rbufsize - ret) 
		     << " bytes to WAV file" << endl;  
	    osize -= rnum;
	} while (osize > 0);

	wave.writeHeader();
    }
    else
    {
	// get normalisation factor if desired:
	double nf = getNormFact(num, data);

	// open output file
	FILE* fpo = fopen(out.c_str(), "w");
	if (!fpo) throw ErrorSys("Cannot open " + out);

	// print comment line:
	fprintf(fpo, "# num | xcorr\n");
	// print entries:
	for (unsigned int i = 0; i < num; ++i)
	    fprintf(fpo, "%d\t%f\n", i, data[i] * nf);
	fclose(fpo);
    }
}


static void process(const string& ref, const string& in, const string& out)
{
    bool autocorr = (ref == in);

    //---------------------------------------------------
    // read the data from the 1. input file (reference):
    //---------------------------------------------------
    uint32 numRef = 0;  // number of samples in 1. file
    uint32 sbRef = 0;   // number of bits in sample
    uint32 sr = 0;       // sample rate in Hz
    int32* data = read_file(ref, numRef, sr, sbRef);
    corr_ref_t* dataRef = new corr_ref_t[numRef];
    if (!dataRef) throw ErrorSysOutOfMem();
    for (uint32 i = 0; i < numRef; ++i)
    {
#ifdef REF_TYPE_BOOL
	dataRef[i] = data[i] > 0;
#else
	if (autocorr) dataRef[i] = data[i];
	else          dataRef[i] = data[i] > 0 ? 1:-1;
#endif
    }

    //--------------------------------------------------
    // read the data from the 2. input file (response):
    //--------------------------------------------------
    uint32 numIn = 0;  // number of samples in 2. file
    uint32 sbIn = 0;   // number of bits in sample
    corr_in_t* dataIn;
    if (!autocorr)
    {
	delete[] data;
	data = read_file(in, numIn, sr, sbIn);
	dataIn = new corr_in_t[numIn];
	if (!dataIn) throw ErrorSysOutOfMem();
    }
    else
    {
	numIn = numRef;
	sbIn = sbRef;
	dataIn = new corr_in_t[numIn];
	if (!dataIn) throw ErrorSysOutOfMem();
    }
    for (uint32 i = 0; i < numIn; ++i)
    {
#ifdef IN_TYPE_BOOL
	dataIn[i] = data[i] > 0;
#else
	dataIn[i] = data[i];
#endif
    }
    delete[] data;

    // use higher sample bit number for output file:
    uint32 sb = (sbRef > sbIn ? sbRef : sbIn);

    //-------------------------------------------------------
    // calculate the cross correlation from the input data:
    //-------------------------------------------------------
    if (debon) cerr << "executing " << (autocorr ? "auto" : "cross")
		    << " correlation" << endl;
    uint32 num = numRef + numIn;        // number of correlated data
    if (cyclic) num = numRef;

    corr_out_t* dout = new corr_out_t[num];
    if (!dout)
    {
	delete[] dataRef;
	if (!autocorr) delete[] dataIn;
	throw ErrorSysOutOfMem();
    }
    correlate(dataRef, numRef, dataIn, numIn, dout, num, cyclic, debon);
    delete[] dataRef;
    if (!autocorr) delete[] dataIn;
    if (debon) calc_stat ("xcorr", dout, num);

    //-----------------------------------------
    // write the results into the output file:
    //-----------------------------------------
    if (debon) cerr << "writing file " << out << endl;

    write_file(out, sr, sb, dout, num);
    delete dout;
}


// --- global functions ---


int main(int argc, char * const argv[])
{
    // loop through command line options:
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'd': ++debon; break;
	case 'C': cyclic = true; break;
	case 'c': channel = atoi(optarg); break;
	case 'h': usage(); exit(0);
	case 'n': normalize = true; break;
	case 'v': print_version(); exit(0);
	default : usage(); exit(1);
	}
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing file" << endl;
	usage();
	exit(1);
    }

    // loop through the remaining arguments: these are files
    vector<string> args;
    for (int i=optind; i < argc; i++)
    {
	args.push_back(argv[i]);
    }

    string in_file_1;
    string in_file_2;
    string out_file;
    switch (args.size())
    {
    case 3:  in_file_1 = args[0];
	     in_file_2 = args[1];
	     out_file = args[2];
	     break;
    case 2:  in_file_1 = in_file_2 = args[0];
	     out_file = args[1];
	     break;
    case 1:  cerr << "Error: missing input file(s)" << endl;
	     usage();
	     exit(1);
	     break;
    case 0:  cerr << "Error: missing input and output files" << endl;
	     usage();
	     exit(1);
    default: cerr << "Error: too many arguments" << endl;
	     usage();
	     exit(1);
    }

    // process the files
    static unsigned int errnum = 0;
    try
    {
	process(in_file_1, in_file_2, out_file); 
    }
    catch (ErrorClass& err)
    {
	++errnum;
	cerr << "Error: " << err.getMsg() << endl;
    }
    catch (exception& err)
    {
	++errnum;
	cerr << "Error: " << err.what() << endl;
    }
    
    exit (errnum ? 1 : 0);
    return  (errnum ? 1 : 0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
